#include <windows.h>
#include <tchar.h>
#define TVTEST_PLUGIN_CLASS_IMPLEMENT
#include "TVTestPlugin.h"
#include "resource.h"


#pragma comment(lib,"winmm.lib")


#define SIGNALTALKER_WINDOW_CLASS TEXT("SignalTalker Window")

#define TIMER_ID		1
#define TIMER_INTERVAL	3000

// obt@TCY
#define BUFFER_SIZE	0x10000


#ifdef _DEBUG
#define TRACE ::OutputDebugString
#else
#define TRACE (void)
#endif


// vOCNX
class CSignalTalker : public TVTest::CTVTestPlugin
{
	enum {
		VOICE_ZERO,
		VOICE_ICHI,
		VOICE_NI,
		VOICE_SAN,
		VOICE_YON,
		VOICE_GO,
		VOICE_ROKU,
		VOICE_NANA,
		VOICE_HACHI,
		VOICE_KYU,
		VOICE_JYU,
		VOICE_NIJYU,
		VOICE_SANJYU,
		VOICE_YONJYU,
		VOICE_GOJYU,
		VOICE_ROKUJYU,
		VOICE_NANAJYU,
		VOICE_HACHIJYU,
		VOICE_KYUJYU,
		NUM_VOICES
	};
	struct {
		void *pData;
		SIZE_T Size;
	} m_VoiceList[NUM_VOICES];
	bool m_fVoiceLoaded;
	HWND m_hwnd;
	bool m_fEnabled;
	bool m_fUsePlaySound;
	float m_SignalLevel;
	HANDLE m_hTalkThread;
	HWAVEOUT m_hWaveOut;
	WAVEHDR m_WaveHdr;
	BYTE *m_pBuffer;

	class CCriticalLock
	{
		CRITICAL_SECTION m_CriticalSection;
	public:
		CCriticalLock() {
			::InitializeCriticalSection(&m_CriticalSection);
		}
		~CCriticalLock() {
			::DeleteCriticalSection(&m_CriticalSection);
		}
		void Lock(void) {
			::EnterCriticalSection(&m_CriticalSection);
		}
		void Unlock(void) {
			::LeaveCriticalSection(&m_CriticalSection);
		}
	};

	class CBlockLock
	{
		CCriticalLock *m_pCriticalLock;
	public:
		CBlockLock(CCriticalLock *pCriticalLock) : m_pCriticalLock(pCriticalLock) {
			m_pCriticalLock->Lock();
		}
		~CBlockLock() {
			m_pCriticalLock->Unlock();
		}
	};

	CCriticalLock m_WaveOutLock;
	static LRESULT CALLBACK EventCallback(UINT Event,LPARAM lParam1,LPARAM lParam2,void *pClientData);
	static CSignalTalker *GetThis(HWND hwnd);
	static LRESULT CALLBACK WndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam);
	void SetVolume(HWAVEOUT hwo,bool fUp);
	bool PlayVoice(LPWAVEFORMATEX pwfx,SIZE_T DataSize);
	void CloseWaveOut();
	static DWORD WINAPI TalkProc(LPVOID lpParameter);

public:
	CSignalTalker();
	virtual bool GetPluginInfo(TVTest::PluginInfo *pInfo);
	virtual bool Initialize();
	virtual bool Finalize();
};


CSignalTalker::CSignalTalker()
{
	for (int i=0;i<NUM_VOICES;i++) {
		m_VoiceList[i].pData=NULL;
		m_VoiceList[i].Size=0;
	}
	m_fVoiceLoaded=false;
	m_hwnd=NULL;
	m_fEnabled=false;
	m_fUsePlaySound=true;
	m_SignalLevel=0.0f;
	m_hTalkThread=NULL;
	m_hWaveOut=NULL;
	m_pBuffer=NULL;
}


bool CSignalTalker::GetPluginInfo(TVTest::PluginInfo *pInfo)
{
	// vOC̏Ԃ
	pInfo->Type           = TVTest::PLUGIN_TYPE_NORMAL;
	pInfo->Flags          = TVTest::PLUGIN_FLAG_HASSETTINGS;
	pInfo->pszPluginName  = L"SignalTalker";
	pInfo->pszCopyright   = L"Public Domain";
	pInfo->pszDescription = L"Mxǂݏグ";
	return true;
}


bool CSignalTalker::Initialize()
{
	// 

	WNDCLASS wc;

	wc.style=CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=g_hinstDLL;
	wc.hIcon=NULL;
	wc.hCursor=::LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=::CreateSolidBrush(RGB(0,0,0));
	wc.lpszMenuName=NULL;
	wc.lpszClassName=SIGNALTALKER_WINDOW_CLASS;
	if (::RegisterClass(&wc)==0)
		return false;

	m_hwnd=::CreateWindowEx(0,SIGNALTALKER_WINDOW_CLASS,TEXT("SignalTalker"),
							WS_OVERLAPPEDWINDOW,
							0,0,320,240,
							m_pApp->GetAppWindow(),NULL,g_hinstDLL,this);
	if (m_hwnd==NULL)
		return false;

	// CxgR[obN֐o^
	m_pApp->SetEventCallback(EventCallback,this);

	return true;
}


bool CSignalTalker::Finalize()
{
	// I

	::DestroyWindow(m_hwnd);

	return true;
}


// CxgR[obN֐
// CxgNƌĂ΂
LRESULT CALLBACK CSignalTalker::EventCallback(UINT Event,LPARAM lParam1,LPARAM lParam2,void *pClientData)
{
	CSignalTalker *pThis=static_cast<CSignalTalker*>(pClientData);

	switch (Event) {
	case TVTest::EVENT_PLUGINENABLE:
		// vOC̗LԂω
		pThis->m_fEnabled=lParam1!=0;
		if (pThis->m_fEnabled) {
			if (!pThis->m_fVoiceLoaded) {
				pThis->m_fVoiceLoaded=true;
				for (int i=0;i<NUM_VOICES;i++) {
					if (pThis->m_VoiceList[i].pData==NULL) {
						HRSRC hres;

						hres=::FindResource(g_hinstDLL,MAKEINTRESOURCE(IDW_ZERO+i),TEXT("WAVE"));
						if (hres!=NULL) {
							HGLOBAL hGlobal;

							pThis->m_VoiceList[i].Size=::SizeofResource(g_hinstDLL,hres);
							hGlobal=::LoadResource(g_hinstDLL,hres);
							pThis->m_VoiceList[i].pData=::LockResource(hGlobal);
						} else {
							pThis->m_fVoiceLoaded=false;
						}
					}
				}
			}
			::ShowWindow(pThis->m_hwnd,SW_SHOW);
			::SetTimer(pThis->m_hwnd,TIMER_ID,TIMER_INTERVAL,NULL);
		} else {
			::KillTimer(pThis->m_hwnd,TIMER_ID);
			::ShowWindow(pThis->m_hwnd,SW_HIDE);
			pThis->CloseWaveOut();
			pThis->m_SignalLevel=0.0f;
		}
		return TRUE;
	}
	return 0;
}


CSignalTalker *CSignalTalker::GetThis(HWND hwnd)
{
	return reinterpret_cast<CSignalTalker*>(::GetWindowLongPtr(hwnd,GWLP_USERDATA));
}


LRESULT CALLBACK CSignalTalker::WndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	switch (uMsg) {
	case WM_CREATE:
		{
			LPCREATESTRUCT pcs=reinterpret_cast<LPCREATESTRUCT>(lParam);
			CSignalTalker *pThis=static_cast<CSignalTalker*>(pcs->lpCreateParams);

			::SetWindowLongPtr(hwnd,GWLP_USERDATA,reinterpret_cast<LONG_PTR>(pThis));
		}
		return 0;

	case WM_PAINT:
		{
			CSignalTalker *pThis=GetThis(hwnd);
			PAINTSTRUCT ps;
			LOGFONT lf;
			HFONT hfont,hfontOld;
			RECT rc;
			int Level;
			TCHAR szText[16];
			COLORREF crOldTextColor;
			int OldBkMode;

			::BeginPaint(hwnd,&ps);
			::GetObject(::GetStockObject(DEFAULT_GUI_FONT),sizeof(LOGFONT),&lf);
			::GetClientRect(hwnd,&rc);
			lf.lfHeight=min(rc.bottom,rc.right/3);
			hfont=::CreateFontIndirect(&lf);
			hfontOld=static_cast<HFONT>(::SelectObject(ps.hdc,hfont));
			crOldTextColor=::SetTextColor(ps.hdc,RGB(255,255,255));
			OldBkMode=::SetBkMode(ps.hdc,TRANSPARENT);
			Level=(int)(pThis->m_SignalLevel*100.0f);
			::wsprintf(szText,TEXT("%d.%02d"),Level/100,Level%100);
			::DrawText(ps.hdc,szText,-1,&rc,DT_CENTER | DT_VCENTER | DT_SINGLELINE);
			::SetBkMode(ps.hdc,OldBkMode);
			::SetTextColor(ps.hdc,crOldTextColor);
			::SelectObject(ps.hdc,hfontOld);
			::DeleteObject(hfont);
			::EndPaint(hwnd,&ps);
		}
		return 0;

	case WM_TIMER:
		{
			CSignalTalker *pThis=GetThis(hwnd);
			TVTest::StatusInfo Status;

			if (pThis->m_hTalkThread) {
				::WaitForSingleObject(pThis->m_hTalkThread,INFINITE);
				::CloseHandle(pThis->m_hTalkThread);
				pThis->m_hTalkThread=NULL;
			}
			if (pThis->m_pApp->GetStatus(&Status)) {
				pThis->m_SignalLevel=Status.SignalLevel;
			} else {
				pThis->m_SignalLevel=0.0f;
			}
			pThis->m_hTalkThread=::CreateThread(NULL,0,TalkProc,pThis,0,NULL);
			::InvalidateRect(hwnd,NULL,TRUE);
		}
		return 0;

	case WM_KEYDOWN:
		if (wParam==VK_ESCAPE) {
			::SendMessage(hwnd,WM_SYSCOMMAND,SC_CLOSE,0);
			return 0;
		}
		break;

	case WM_MOUSEWHEEL:
		if ((SHORT)HIWORD(wParam)>0)
			uMsg=WM_LBUTTONDOWN;
		else
			uMsg=WM_RBUTTONDOWN;
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		{
			CSignalTalker *pThis=GetThis(hwnd);
			CBlockLock Lock(&pThis->m_WaveOutLock);

			if (pThis->m_fUsePlaySound) {
				HWAVEOUT hwo;
				WAVEFORMATEX wfx;

				wfx.wFormatTag=WAVE_FORMAT_PCM;
				wfx.nChannels=1;
				wfx.nSamplesPerSec=22050;
				wfx.nAvgBytesPerSec=22050;
				wfx.nBlockAlign=1;
				wfx.wBitsPerSample=8;
				if (::waveOutOpen(&hwo,WAVE_MAPPER,&wfx,0,0,CALLBACK_NULL)==MMSYSERR_NOERROR) {
					pThis->SetVolume(hwo,uMsg==WM_LBUTTONDOWN);
					::waveOutClose(hwo);
				}
			} else if (pThis->m_hWaveOut!=NULL) {
				pThis->SetVolume(pThis->m_hWaveOut,uMsg==WM_LBUTTONDOWN);
			}
		}
		return 0;

	case WM_SYSCOMMAND:
		if ((wParam&0xFFFFFFF0)==SC_CLOSE) {
			CSignalTalker *pThis=GetThis(hwnd);

			pThis->m_pApp->EnablePlugin(false);
			return 0;
		}
		break;

	case WM_DESTROY:
		{
			CSignalTalker *pThis=GetThis(hwnd);

			if (pThis->m_fEnabled)
				::KillTimer(hwnd,TIMER_ID);	// ʂɂȂĂ...
			pThis->CloseWaveOut();
		}
		return 0;
	}
	return ::DefWindowProc(hwnd,uMsg,wParam,lParam);
}


void CSignalTalker::SetVolume(HWAVEOUT hwo,bool fUp)
{
	DWORD Volume;

	if (::waveOutGetVolume(hwo,&Volume)==MMSYSERR_NOERROR) {
		DWORD Left=LOWORD(Volume),Right=HIWORD(Volume);

		if (fUp) {
			if (Left+1024>0xFFFF)
				Left=0xFFFF;
			else
				Left+=1024;
			if (Right+1024>0xFFFF)
				Right=0xFFFF;
			else
				Right+=1024;
		} else {
			if (Left<1024)
				Left=0;
			else
				Left-=1024;
			if (Right<1024)
				Right=0;
			else
				Right-=1024;
		}
		::waveOutSetVolume(hwo,Left|(Right<<16));
	}
}


bool CSignalTalker::PlayVoice(LPWAVEFORMATEX pwfx,SIZE_T DataSize)
{
	CBlockLock Lock(&m_WaveOutLock);

	if (m_hWaveOut==NULL) {
		if (::waveOutOpen(&m_hWaveOut,WAVE_MAPPER,pwfx,(UINT)m_hwnd,0,CALLBACK_WINDOW)!=0) {
			m_hWaveOut=NULL;
			TRACE(TEXT("Error : waveOutOpen()\n"));
			return false;
		}
		::ZeroMemory(&m_WaveHdr,sizeof(WAVEHDR));
		m_WaveHdr.lpData=(LPSTR)m_pBuffer;
		m_WaveHdr.dwBufferLength=BUFFER_SIZE;
		m_WaveHdr.dwFlags=WHDR_DONE;
		m_WaveHdr.dwLoops=0;
		if (::waveOutPrepareHeader(m_hWaveOut,&m_WaveHdr,sizeof(WAVEHDR))!=0) {
			::waveOutClose(m_hWaveOut);
			m_hWaveOut=NULL;
			TRACE(TEXT("Error : waveOutPrepareHeader()\n"));
			return false;
		}
	}
	m_WaveHdr.dwBufferLength=DataSize;
	if (::waveOutWrite(m_hWaveOut,&m_WaveHdr,sizeof(WAVEHDR))!=0) {
		::waveOutReset(m_hWaveOut);
		::waveOutUnprepareHeader(m_hWaveOut,&m_WaveHdr,sizeof(WAVEHDR));
		::waveOutClose(m_hWaveOut);
		m_hWaveOut=NULL;
		TRACE(TEXT("Error : waveOutWrite()\n"));
		return false;
	}
	return true;
}


void CSignalTalker::CloseWaveOut()
{
	if (m_hTalkThread!=NULL) {
		::WaitForSingleObject(m_hTalkThread,INFINITE);
		::CloseHandle(m_hTalkThread);
		m_hTalkThread=NULL;
	}
	if (m_hWaveOut!=NULL) {
		::waveOutReset(m_hWaveOut);
		::waveOutUnprepareHeader(m_hWaveOut,&m_WaveHdr,sizeof(WAVEHDR));
		::waveOutClose(m_hWaveOut);
		m_hWaveOut=NULL;
	}
	if (m_pBuffer!=NULL) {
		delete [] m_pBuffer;
		m_pBuffer=NULL;
	}
}


DWORD WINAPI CSignalTalker::TalkProc(LPVOID lpParameter)
{
	CSignalTalker *pThis=static_cast<CSignalTalker*>(lpParameter);
	int Level=(int)(pThis->m_SignalLevel*100.0f)/100;
	WAVEFORMATEX *pwfx;
	BYTE *pData;
	DWORD SampleSize,Size;

	if (Level>99)
		return 0;
	if (pThis->m_pBuffer==NULL)
		pThis->m_pBuffer=new BYTE[BUFFER_SIZE];
	SampleSize=0;
	if (Level>=10 && pThis->m_VoiceList[VOICE_JYU+Level/10-1].pData!=NULL) {
		// 蔲ŃtH[}bgߑł
		pData=static_cast<BYTE*>(pThis->m_VoiceList[VOICE_JYU+Level/10-1].pData);
		pData+=20;	// "RIFF" + File size + "WAVE" + "fmt " + Format size
		pwfx=reinterpret_cast<WAVEFORMATEX*>(pData);
		pData+=sizeof(PCMWAVEFORMAT);
		if (pData[0]!='d' || pData[1]!='a' || pData[2]!='t' || pData[3]!='a')
			return 0;
		pData+=4;
		SampleSize=pData[0] | (pData[1]<<8) | (pData[2]<<16) | (pData[3]<<24);
		if (pThis->m_fUsePlaySound) {
			SampleSize+=20+sizeof(PCMWAVEFORMAT)+8;
			::CopyMemory(pThis->m_pBuffer,
						 pThis->m_VoiceList[VOICE_JYU+Level/10-1].pData,
						 SampleSize);
		} else {
			::CopyMemory(pThis->m_pBuffer,pData+4,SampleSize);
		}
	}
	if ((Level==0 || Level%10!=0)
			&& pThis->m_VoiceList[VOICE_ZERO+Level%10].pData!=NULL) {
		pData=static_cast<BYTE*>(pThis->m_VoiceList[VOICE_ZERO+Level%10].pData);
		pData+=20;	// "RIFF" + File size + "WAVE" + "fmt " + Format size
		pwfx=reinterpret_cast<WAVEFORMATEX*>(pData);
		pData+=sizeof(PCMWAVEFORMAT);
		if (pData[0]!='d' || pData[1]!='a' || pData[2]!='t' || pData[3]!='a')
			return 0;
		pData+=4;
		Size=pData[0] | (pData[1]<<8) | (pData[2]<<16) | (pData[3]<<24);
		if (pThis->m_fUsePlaySound) {
			if (SampleSize==0) {
				::CopyMemory(pThis->m_pBuffer+SampleSize,
							 pThis->m_VoiceList[VOICE_ZERO+Level%10].pData,
							 20+sizeof(PCMWAVEFORMAT)+8);
				SampleSize=20+sizeof(PCMWAVEFORMAT)+8;
			} else {
				*(DWORD*)(pThis->m_pBuffer+(20+sizeof(PCMWAVEFORMAT)+4))+=Size;
			}
		}
		::CopyMemory(pThis->m_pBuffer+SampleSize,pData+4,Size);
		SampleSize+=Size;
	}
	if (SampleSize>0) {
		if (pThis->m_fUsePlaySound) {
			::PlaySound((LPCTSTR)pThis->m_pBuffer,NULL,
						SND_MEMORY | SND_ASYNC | SND_NODEFAULT);
		} else {
			pThis->PlayVoice(pwfx,SampleSize);
		}
	}
	return 0;
}




TVTest::CTVTestPlugin *CreatePluginClass()
{
	return new CSignalTalker;
}
